import json
from typing import Any, Generator
from unittest.mock import patch
from uuid import uuid4

import pytest

from pasqal_cloud import SDK
from pasqal_cloud.errors import DeviceSpecsFetchingError
from tests.test_doubles.authentication import FakeAuth0AuthenticationSuccess


class TestDeviceSpecs:
    @pytest.fixture(autouse=True)
    @patch("sdk.client.Auth0TokenProvider", FakeAuth0AuthenticationSuccess)
    def _init_sdk(self):
        self.sdk = SDK(
            username="me@test.com", password="password", project_id=str(uuid4())
        )

    @pytest.fixture(autouse=True)
    def _mock_sleep(self):
        """
        This fixture overrides sleeps, so tests don't need to wait for
        the total elapsed time.
        """
        with patch("time.sleep"):
            yield

    @pytest.mark.usefixtures("mock_request")
    def test_get_device_specs_success(self):
        device_specs_dict = self.sdk.get_device_specs_dict()
        assert isinstance(device_specs_dict, dict)
        specs = device_specs_dict["FRESNEL"]
        json.loads(specs)

    def test_get_device_specs_error(
        self, mock_request_exception: Generator[Any, Any, None]
    ):
        with pytest.raises(DeviceSpecsFetchingError):
            _ = self.sdk.get_device_specs_dict()
        assert mock_request_exception.last_request.method == "GET"
        assert (
            mock_request_exception.last_request.url
            == f"{self.sdk._client.endpoints.core}/api/v1/devices/specs"
        )



import json
from typing import Optional

from requests import ConnectionError, HTTPError, Response


class ExceptionWithResponseContext(BaseException):
    def __init__(self, msg: str, e: Optional[HTTPError] = None):
        if not e:
            return super().__init__(msg)

        if e.response is None:
            return super().__init__(msg)

        resp: Response = e.response

        if not resp.content:
            return super().__init__(msg)
        data = resp.json()

        code = data.get("code", 0)

        if data.get("data", ""):
            description = data["data"].get("description", data.get("message"))
        else:
            description = data.get("message", "")

        return super().__init__(
            f"{msg}: {code}: {description}\nDetails: {json.dumps(data, indent=2)}"
        )


class BatchException(ExceptionWithResponseContext):
    """
    Base Exception class for batches
    """


class BatchCreationError(BatchException):
    """
    Exception class when batch creation failed
    """

    def __init__(self, e: HTTPError) -> None:
        super().__init__("Batch creation failed", e)


class BatchFetchingError(BatchException):
    """
    Exception class raised when batch fetching failed.
    """

    def __init__(self, e: HTTPError) -> None:
        super().__init__("Batch fetching failed", e)


class BatchCancellingError(BatchException):
    """
    Exception class raised when batch cancelling failed.
    """

    def __init__(self, e: HTTPError) -> None:
        super().__init__("Batch cancelling failed", e)


class BatchSetCompleteError(BatchException):
    """
    Exception class raised when setting batch to complete failed.
    """

    def __init__(self, e: HTTPError) -> None:
        super().__init__("Unable to set batch as complete.", e)


class RebatchError(BatchException):
    """Exception class raised when batch retry failed."""

    def __init__(self, e: HTTPError) -> None:
        super().__init__("Batch retry failed", e)


class JobException(ExceptionWithResponseContext):
    """
    Base Exception class for jobs.
    """


class JobCreationError(JobException):
    """
    Exception class raised when job creation failed.
    """

    def __init__(self, e: HTTPError) -> None:
        super().__init__("Job creation failed", e)


class JobFetchingError(JobException):
    """
    Exception class raised when job fetching failed.
    """

    def __init__(self, e: HTTPError) -> None:
        super().__init__("Job fetching failed", e)


class JobCancellingError(JobException):
    """
    Exception class raised when job cancelling failed.
    """

    def __init__(self, e: HTTPError) -> None:
        super().__init__("Job cancelling failed", e)


class JobRetryError(Exception):
    """
    Exception class raised when job retry failed.
    """

    def __init__(self) -> None:
        super().__init__("Job retry failed.")


class WorkloadException(ExceptionWithResponseContext):
    """
    Base exception class for workloads.
    """


class WorkloadFetchingError(WorkloadException):
    """
    Exception class raised when workload fetching failed.
    """

    def __init__(self, e: HTTPError) -> None:
        super().__init__("Workload fetching failed", e)


class WorkloadCreationError(WorkloadException):
    """
    Exception class raised when workload creation failed.
    """

    def __init__(self, e: HTTPError) -> None:
        super().__init__("Job creation failed", e)


class WorkloadCancellingError(WorkloadException):
    """
    Exception class raised when cancelling workload failed.
    """

    def __init__(self, e: HTTPError) -> None:
        super().__init__("Workload cancelling failed", e)


class WorkloadResultsDownloadError(WorkloadException):
    """
    Exception class raised when failed to download results.
    """

    def __init__(self, e: HTTPError) -> None:
        super().__init__("Workload results download failed.", e)


class WorkloadResultsConnectionError(BaseException):
    """
    Exception class raised when failed to download results
    for a connection error.
    """

    def __init__(self, e: ConnectionError) -> None:
        super().__init__("Workload results download failed from connection error.", e)


class WorkloadResultsDecodeError(WorkloadException):
    """
    Exception class raised when download results succeeded but decoding failed.
    """

    def __init__(self) -> None:
        super().__init__("Workload results decoding failed.")


class InvalidWorkloadResultsFormatError(WorkloadException):
    """
    Exception class raised when download results succeeded but format
    is not as expected.
    """

    def __init__(self, result_type: type) -> None:
        super().__init__(f"Workload results should be dict but received {result_type}")


class DeviceSpecsException(BaseException):
    """
    Base Exception class for device specs
    """


class DeviceSpecsFetchingError(DeviceSpecsException):
    """
    Exception class raised when fetching of device specs failed.
    """

    def __init__(self, e: HTTPError) -> None:
        super().__init__("Device specs fetching failed", e)



from __future__ import annotations

import re

import pytest

from pasqal_cloud.device import EmuFreeConfig
from pasqal_cloud.device.configuration.base_config import (
    BaseConfig,
    INVALID_KEY_ERROR_MSG,
    INVALID_RESULT_TYPES,
    InvalidConfiguration,
)
from pasqal_cloud.device.configuration.emu_tn import (
    DT_VALUE_NOT_VALID,
    EmuTNConfig,
    PRECISION_NOT_VALID,
)
from pasqal_cloud.device.configuration.result_type import ResultType


@pytest.mark.parametrize(
    ("config", "expected"),
    [
        (
            EmuTNConfig(
                dt=10.0,
                extra_config={
                    "extra": "parameter",
                    "extra_dict": {"key": "value"},
                },
            ),
            {
                "strict_validation": False,
                "dt": 10.0,
                "precision": "normal",
                "max_bond_dim": 500,
                "extra": "parameter",
                "extra_dict": {"key": "value"},
                "result_types": None,
            },
        ),
        (
            EmuTNConfig(result_types=[ResultType.COUNTER]),
            {
                "strict_validation": False,
                "dt": 10.0,
                "precision": "normal",
                "max_bond_dim": 500,
                "result_types": [ResultType.COUNTER],
            },
        ),
        (
            EmuFreeConfig(),
            {
                "strict_validation": False,
                "with_noise": False,
                "result_types": None,
            },
        ),
        (
            EmuFreeConfig(with_noise=True),
            {
                "strict_validation": False,
                "with_noise": True,
                "result_types": None,
            },
        ),
        (
            BaseConfig(),
            {
                "strict_validation": False,
                "result_types": None,
            },
        ),
        (
            BaseConfig(strict_validation=True),
            {
                "strict_validation": True,
                "result_types": None,
            },
        ),
        (
            BaseConfig(extra_config={"extra": "parameter"}),
            {
                "strict_validation": False,
                "extra": "parameter",
                "result_types": None,
            },
        ),
    ],
)
def test_configuration_to_dict(config: BaseConfig, expected: dict):
    assert config.to_dict() == expected


@pytest.mark.parametrize(
    ("config_class", "expected", "config"),
    [
        (
            EmuTNConfig,
            EmuTNConfig(
                dt=0.5,
                extra_config={"extra": "parameter", "extra_dict": {"key": "value"}},
            ),
            {
                "strict_validation": False,
                "dt": 0.5,
                "precision": "normal",
                "extra": "parameter",
                "extra_dict": {"key": "value"},
            },
        ),
        (
            EmuTNConfig,
            EmuTNConfig(),
            {"strict_validation": False, "dt": 10.0, "precision": "normal"},
        ),
        (
            EmuTNConfig,
            EmuTNConfig(strict_validation=True),
            {"strict_validation": True, "dt": 10.0, "precision": "normal"},
        ),
        (
            EmuFreeConfig,
            EmuFreeConfig(),
            {"strict_validation": False, "with_noise": False},
        ),
        (
            EmuFreeConfig,
            EmuFreeConfig(with_noise=True),
            {"strict_validation": False, "with_noise": True},
        ),
        (
            EmuFreeConfig,
            EmuFreeConfig(strict_validation=True),
            {"strict_validation": True, "with_noise": False},
        ),
        (
            BaseConfig,
            BaseConfig(),
            {
                "strict_validation": False,
            },
        ),
        (
            BaseConfig,
            BaseConfig(extra_config={"extra": "parameter"}),
            {"strict_validation": False, "extra": "parameter"},
        ),
    ],
)
def test_configuration_from_dict(
    config_class: type, expected: BaseConfig, config: dict
):
    assert config_class.from_dict(config) == expected


@pytest.mark.parametrize(
    ("config", "extra_config", "expected"),
    [
        (EmuTNConfig(), {"dt": 10.0}, INVALID_KEY_ERROR_MSG.format("dt")),
        (EmuTNConfig(dt=-1), None, DT_VALUE_NOT_VALID.format(-1)),
        (
            EmuTNConfig(precision="nonsense"),
            None,
            PRECISION_NOT_VALID.format("nonsense"),
        ),
        (
            EmuTNConfig(result_types=[ResultType.EXPECTATION]),
            None,
            INVALID_RESULT_TYPES.format(
                [ResultType.EXPECTATION], EmuTNConfig().allowed_result_types
            ),
        ),
    ],
)
def test_wrong_configuration(config, extra_config, expected):
    config.extra_config = extra_config
    with pytest.raises(InvalidConfiguration, match=re.escape(expected)):
        config.to_dict()



import json

import pytest
from requests import HTTPError, Response

from pasqal_cloud.errors import ExceptionWithResponseContext


def test_exception_for_response_with_description():
    """
    When we catch an exception with a populated Response object,
    the error message should extract data from the object
    and format the message. Then we assert it's as expected
    with an f-string pattern.
    """
    response = Response()
    response.status_code = 400
    description = "Unprocessable entity."
    data = {
        "code": response.status_code,
        "data": {"description": description, "some": "data"},
    }
    response._content = json.dumps(data).encode("utf-8")
    error = HTTPError(response=response)
    with pytest.raises(
        ExceptionWithResponseContext,
        match=(
            f"some message: {response.status_code}: {description}\nDetails: "
            + f"{json.dumps(data, indent=2)}"
        ),
    ):
        raise ExceptionWithResponseContext("some message", error)


def test_exception_for_response_without_description():
    """
    When we catch an exception with a populated Response object,
    the error message should extract data from the object
    and format the message. Then we assert it's as expected
    with an f-string pattern.
    """
    response = Response()
    response.status_code = 400
    data = {"message": "message"}
    response._content = json.dumps(data).encode("utf-8")
    error = HTTPError(response=response)
    with pytest.raises(
        ExceptionWithResponseContext,
        match=(f"some message: 0: message\nDetails: {json.dumps(data, indent=2)}"),
    ):
        raise ExceptionWithResponseContext("some message", error)


def test_exception_with_response_without_data():
    """
    When we catch an exception with a unpopulated Response object,
    the error message should extract the data it contains
    and format the message. Then we assert it's as expected.
    """
    response = Response()
    response.status_code = 400
    error = HTTPError(response=response)
    with pytest.raises(ExceptionWithResponseContext, match="some message"):
        raise ExceptionWithResponseContext("some message", error)


def test_exception_without_response():
    with pytest.raises(ExceptionWithResponseContext, match="some message"):
        raise ExceptionWithResponseContext("some message")



# Copyright 2020 Pasqal Cloud Services development team
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from sys import version_info
from typing import Any, Dict, List, Optional

if version_info[:2] >= (3, 8):
    from typing import TypedDict
else:
    from typing_extensions import TypedDict


class JobResult(TypedDict):
    raw: List[str]
    counter: Dict[str, Any]


class PaginationType(TypedDict):
    total: int
    offset: int
    end: int


class JSendPayload(TypedDict):
    status: str
    code: int
    message: str
    data: Any
    pagination: Optional[PaginationType]



from pasqal_cloud.device.configuration import BaseConfig, EmuFreeConfig, EmuTNConfig
from pasqal_cloud.device.emulator_types import EmulatorType

__all__ = [
    "BaseConfig",
    "EmuFreeConfig",
    "EmuTNConfig",
    "EmulatorType",
]



# Copyright 2020 Pasqal Cloud Services development team
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from dataclasses import dataclass
from sys import version_info

if version_info[:2] >= (3, 8):
    from typing import Final
else:
    from typing_extensions import Final  # type: ignore[assignment]


# ---- Endpoints ----

# Prod Core API URL
CORE_API_URL: Final[str] = "https://apis.pasqal.cloud/core-fast"

PREPROD_CORE_API_URL: Final[str] = "https://apis.preprod.pasqal.cloud/core-fast"

DEV_CORE_API_URL: Final[str] = "https://apis.dev.pasqal.cloud/core-fast"


@dataclass
class Endpoints:
    core: str = CORE_API_URL


# ---- Auth0 ----

# Prod Auth0 configuration
AUTH0_DOMAIN: Final[str] = "authenticate.pasqal.cloud"
PUBLIC_CLIENT_ID: Final[str] = "PeZvo7Atx7IVv3iel59asJSb4Ig7vuSB"
AUDIENCE: Final[str] = "https://apis.pasqal.cloud/account/api/v1"

PREPROD_AUTH0_DOMAIN: Final[str] = "authenticate.preprod.pasqal.cloud"
PREPROD_PUBLIC_CLIENT_ID: Final[str] = "2l6A2ldvwJE5sdkghu40BTYLm7sSUAv9"
PREPROD_AUDIENCE: Final[str] = "https://apis.preprod.pasqal.cloud/account/api/v1"

DEV_AUTH0_DOMAIN: Final[str] = "authenticate.dev.pasqal.cloud"
DEV_PUBLIC_CLIENT_ID: Final[str] = "5QtfSu1UV118Iz6By6IJRSNoDrLbAiOv"
DEV_AUDIENCE: Final[str] = "https://apis.dev.pasqal.cloud/account/api/v1"

REALM: Final[str] = "pcs-users"


@dataclass
class Auth0Conf:
    domain: str = AUTH0_DOMAIN
    public_client_id: str = PUBLIC_CLIENT_ID
    audience: str = AUDIENCE
    realm: str = REALM


# ---- Pre-filled environment configuration ----


PASQAL_ENDPOINTS = {
    "prod": Endpoints(core=CORE_API_URL),
    "preprod": Endpoints(core=PREPROD_CORE_API_URL),
    "dev": Endpoints(core=DEV_CORE_API_URL),
}

AUTH0_CONFIG = {
    "prod": Auth0Conf(
        domain=AUTH0_DOMAIN,
        public_client_id=PUBLIC_CLIENT_ID,
        audience=AUDIENCE,
        realm=REALM,
    ),
    "preprod": Auth0Conf(
        domain=PREPROD_AUTH0_DOMAIN,
        public_client_id=PREPROD_PUBLIC_CLIENT_ID,
        audience=PREPROD_AUDIENCE,
        realm=REALM,
    ),
    "dev": Auth0Conf(
        domain=DEV_AUTH0_DOMAIN,
        public_client_id=DEV_PUBLIC_CLIENT_ID,
        audience=DEV_AUDIENCE,
        realm=REALM,
    ),
}



import json
from typing import Any, Generator
from unittest.mock import patch
from uuid import UUID, uuid4

import pytest
import requests

from pasqal_cloud import SDK, Workload
from pasqal_cloud.errors import (
    WorkloadCancellingError,
    WorkloadCreationError,
    WorkloadFetchingError,
    WorkloadResultsConnectionError,
    WorkloadResultsDecodeError,
)
from tests.test_doubles.authentication import FakeAuth0AuthenticationSuccess


class TestWorkload:
    @pytest.fixture()
    def workload_with_link_id(self) -> str:
        return str(UUID(int=0x2))

    @pytest.fixture()
    def workload_with_invalid_link_id(self) -> str:
        return str(UUID(int=0x3))

    @pytest.fixture(autouse=True)
    @patch(
        "pasqal_cloud.client.Auth0TokenProvider",
        FakeAuth0AuthenticationSuccess,
    )
    def _init_sdk(self):
        self.sdk = SDK(
            username="me@test.com",
            password="password",
            project_id=str(uuid4()),
        )
        self.workload_id = "00000000-0000-0000-0000-000000000001"
        self.backend = "backend_test"
        self.workload_type = "workload_type_test"
        self.config = {"test1": "test1", "test2": 2}
        self.workload_result = {"1001": 12, "0110": 35, "1111": 1}

    @pytest.fixture(autouse=True)
    def _mock_sleep(self):
        """
        This fixture overrides sleeps, so tests don't need to wait for
        the entire duration of a sleep command.
        """
        with patch("time.sleep"):
            yield

    def test_create_workload(self, mock_request: Generator[Any, Any, None]):
        workload = self.sdk.create_workload(
            backend=self.backend,
            workload_type=self.workload_type,
            config=self.config,
        )
        assert workload.id == self.workload_id
        assert workload.backend == self.backend
        assert workload.workload_type == self.workload_type
        assert workload.config == self.config
        assert (
            mock_request.last_request.url == f"{self.sdk._client.endpoints.core}"
            f"/api/v1/workloads"
        )
        assert mock_request.last_request.method == "POST"

    def test_create_workload_error(
        self, mock_request_exception: Generator[Any, Any, None]
    ):
        with pytest.raises(WorkloadCreationError):
            _ = self.sdk.create_workload(
                backend=self.backend,
                workload_type=self.workload_type,
                config=self.config,
            )
        assert (
            mock_request_exception.last_request.url
            == f"{self.sdk._client.endpoints.core}"
            f"/api/v1/workloads"
        )
        assert mock_request_exception.last_request.method == "POST"

    def test_create_workload_and_wait(self, mock_request: Generator[Any, Any, None]):
        workload = self.sdk.create_workload(
            backend=self.backend,
            workload_type=self.workload_type,
            config=self.config,
            wait=True,
        )
        assert workload.id == self.workload_id
        assert workload.backend == self.backend
        assert workload.workload_type == self.workload_type
        assert workload.config == self.config
        assert workload.result == self.workload_result
        assert mock_request.last_request.method == "GET"

    def test_get_workload(
        self, mock_request: Generator[Any, Any, None], workload: Workload
    ):
        workload_requested = self.sdk.get_workload(workload.id)
        assert workload_requested.id == self.workload_id
        assert (
            mock_request.last_request.url == f"{self.sdk._client.endpoints.core}"
            f"/api/v2/workloads/{self.workload_id}"
        )

    def test_get_workload_with_link(
        self,
        mock_request: Generator[Any, Any, None],
        workload_with_link_id: Workload,
        result_link_endpoint: str,
    ):
        self.sdk.get_workload(workload_with_link_id)
        assert mock_request.last_request.url == (
            f"{result_link_endpoint}{workload_with_link_id}"
        )

    def test_get_workload_with_invalid_link(
        self,
        workload_with_invalid_link_id: Workload,
        mock_request: Generator[Any, Any, None],
    ):
        with pytest.raises(WorkloadResultsDecodeError):
            self.sdk.get_workload(workload_with_invalid_link_id)
        assert (
            mock_request.last_request.url
            == "http://invalid-link/00000000-0000-0000-0000-000000000003"
        )

    def test_get_workload_error(
        self, mock_request_exception: Generator[Any, Any, None], workload: Workload
    ):
        with pytest.raises(WorkloadFetchingError):
            _ = self.sdk.get_workload(workload.id)
        assert (
            mock_request_exception.last_request.url
            == f"{self.sdk._client.endpoints.core}"
            f"/api/v2/workloads/{self.workload_id}"
        )
        assert mock_request_exception.last_request.method == "GET"

    def test_cancel_workload_self(
        self, mock_request: Generator[Any, Any, None], workload: Workload
    ):
        workload.cancel()
        assert workload.status == "CANCELED"
        assert mock_request.last_request.method == "PUT"
        assert (
            mock_request.last_request.url == f"{self.sdk._client.endpoints.core}"
            f"/api/v1/workloads/{self.workload_id}/cancel"
        )

    def test_cancel_workload_self_error(self, mock_request_exception, workload):
        with pytest.raises(WorkloadCancellingError):
            workload.cancel()
        assert workload.status == "PENDING"
        assert mock_request_exception.last_request.method == "PUT"
        assert (
            mock_request_exception.last_request.url
            == f"{self.sdk._client.endpoints.core}"
            f"/api/v1/workloads/{self.workload_id}/cancel"
        )

    def test_cancel_workload_sdk(self, mock_request: Generator[Any, Any, None]):
        client_rsp = self.sdk.cancel_workload(self.workload_id)
        assert type(client_rsp) == Workload
        assert client_rsp.status == "CANCELED"
        assert mock_request.last_request.method == "PUT"
        assert (
            mock_request.last_request.url == f"{self.sdk._client.endpoints.core}"
            f"/api/v1/workloads/{self.workload_id}/cancel"
        )

    def test_cancel_workload_sdk_error(
        self, mock_request_exception: Generator[Any, Any, None], workload: Workload
    ):
        with pytest.raises(WorkloadCancellingError):
            _ = self.sdk.cancel_workload(self.workload_id)
        assert workload.status == "PENDING"
        assert mock_request_exception.last_request.method == "PUT"
        assert (
            mock_request_exception.last_request.url
            == f"{self.sdk._client.endpoints.core}"
            f"/api/v1/workloads/{self.workload_id}/cancel"
        )

    def test_workload_instantiation_with_extra_field(self, workload: Workload):
        """Instantiating a workload with an extra field should not raise an error.

        This enables us to add new fields in the API response on the workloads endpoint
        without breaking compatibility for users with old versions of the SDK where
        the field is not present in the Batch class.
        """
        workload_dict = workload.model_dump()  # Batch data expected by the SDK
        # We add an extra field to mimick the API exposing new values to the user
        workload_dict["new_field"] = "any_value"

        new_workload = Workload(**workload_dict)  # this should raise no error
        assert (
            new_workload.new_field == "any_value"
        )  # The new value should be stored regardless

    def test_workload_result_raise_connection_error(self, workload: Workload):
        """
        Check that error is raised when an improper url is set.
        """
        workload_dict = workload.model_dump()
        workload_dict.pop("result")
        workload_dict["result_link"] = "http://test.test"
        with pytest.raises(WorkloadResultsConnectionError):
            Workload(**workload_dict)

    def tests_workload_result_set(self, workload: Workload):
        """
        Check that result is set when only result_link is provided.
        """
        workload_dict = workload.model_dump()
        workload_dict.pop("result")
        workload_dict["result_link"] = "http://test.test"
        resp = requests.Response()
        resp._content = json.dumps({"some": "data"}).encode("utf-8")
        with patch("requests.get", return_value=resp):
            res = Workload(**workload_dict)
        assert res.result == {"some": "data"}



# ruff: noqa: F401

import inspect

import pytest

"""
Tests written to verify if the backward compatibility is working between the new name
of the package `pasqal-cloud` and its main folder renamed to `pasqal_cloud` and the
previous package name that was `pasqal-cloud` with the former folder called `sdk`.
"""


def test_verify_import_sdk_is_module():
    with pytest.warns(DeprecationWarning):
        import sdk

        assert inspect.ismodule(sdk)


def test_verify_import_classes_from_sdk():
    with pytest.warns(DeprecationWarning):
        from sdk import Batch, EmulatorType, Endpoints, Job, SDK
        from sdk.device.configuration import BaseConfig, EmuFreeConfig, EmuTNConfig
        from sdk.utils import JSendPayload, StrEnum



from __future__ import annotations

from dataclasses import dataclass

from pasqal_cloud.device.configuration.base_config import (
    BaseConfig,
    InvalidConfiguration,
)

DT_VALUE_NOT_VALID = "dt must be larger than 0. Not {}."
PRECISION_NOT_VALID = "Precision {} not valid. Must be one of 'low', 'normal', 'high'"
MAX_BOND_DIM_NOT_VALID = "Max bond dimension {} must be larger than 0. Not {}."


@dataclass
class EmuTNConfig(BaseConfig):
    """Configuration for the EmuTN device type.

    Args:
        dt (float): The time step in nanoseconds of the simulation. Defaults to 10.0.
        precision (str): The precision of the simulation. Defaults to "normal".
        max_bond_dim (int): The maximum bond dimension of the Matrix Product
          State (MPS). Defaults to 500.
    """

    dt: float = 10.0
    precision: str = "normal"
    max_bond_dim: int = 500

    def _validate(self) -> None:
        if self.dt <= 0:
            raise InvalidConfiguration(DT_VALUE_NOT_VALID.format(self.dt))
        if self.precision not in ["low", "normal", "high"]:
            raise InvalidConfiguration(PRECISION_NOT_VALID.format(self.precision))
        if self.max_bond_dim <= 0:
            raise InvalidConfiguration(MAX_BOND_DIM_NOT_VALID.format(self.max_bond_dim))
        super()._validate()



from pasqal_cloud.device.configuration.base_config import (
    BaseConfig,
    InvalidConfiguration,
)
from pasqal_cloud.device.configuration.emu_free import EmuFreeConfig
from pasqal_cloud.device.configuration.emu_tn import EmuTNConfig

__all__ = [
    "BaseConfig",
    "EmuFreeConfig",
    "EmuTNConfig",
    "InvalidConfiguration",
]



from dataclasses import dataclass

from pasqal_cloud.device.configuration.base_config import BaseConfig


@dataclass
class EmuFreeConfig(BaseConfig):
    """Configuration for the EmuFree device type.

    Args:
        with_noise (bool): Whether to add noise to the simulation. Defaults to False.
    """

    with_noise: bool = False



# ruff: noqa: E402 F403

from sdk import pasqal_sdk_not_updated

pasqal_sdk_not_updated()
from pasqal_cloud.device.configuration import *
from pasqal_cloud.device.configuration.base_config import *
from pasqal_cloud.device.configuration.emu_free import *
from pasqal_cloud.device.configuration.emu_tn import *



# Copyright 2020 Pasqal Cloud Services development team
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from setuptools import find_packages, setup

__version__ = ""
exec(open("pasqal_cloud/_version.py").read())

setup(
    name="pasqal-cloud",
    version=__version__,
    description="Software development kit for Pasqal cloud platform.",
    packages=find_packages(),
    package_data={"pasqal_cloud": ["py.typed"]},
    include_package_data=True,
    long_description=open("README.md").read(),
    long_description_content_type="text/markdown",
    maintainer="Pasqal Cloud Services",
    maintainer_email="pcs@pasqal.io",
    python_requires=">=3.8",
    license="Apache 2.0",
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Programming Language :: Python :: 3",
    ],
    url="https://github.com/pasqal-io/pasqal-cloud",
    install_requires=[
        "auth0-python >= 3.23.1, <4.0.0",
        "requests>=2.25.1, <3.0.0",
        "pyjwt[crypto]>=2.5.0, <3.0.0",
        "pydantic >= 2.6.0, <3.0.0",
    ],
    extras_require={
        "dev": {
            "ruff==0.4.7",
            "mypy==1.10.0",
            "pytest==8.1.1",
            "pytest-cov==5.0.0",
            "types-requests==2.31.0.1",
            "requests-mock==1.12.1",
        }
    },
)



# Copyright 2020 Pasqal Cloud Services development team
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import time
from typing import Any, Dict, List, Optional, Union
from uuid import UUID
from warnings import warn

from requests.exceptions import HTTPError

from pasqal_cloud.authentication import TokenProvider
from pasqal_cloud.batch import Batch, RESULT_POLLING_INTERVAL
from pasqal_cloud.client import Client
from pasqal_cloud.device import BaseConfig, EmulatorType
from pasqal_cloud.endpoints import (
    AUTH0_CONFIG,  # noqa: F401
    Auth0Conf,
    Endpoints,
    PASQAL_ENDPOINTS,  # noqa: F401
)
from pasqal_cloud.errors import (
    BatchCancellingError,
    BatchCreationError,
    BatchFetchingError,
    BatchSetCompleteError,
    DeviceSpecsFetchingError,
    JobCancellingError,
    JobCreationError,
    JobFetchingError,
    RebatchError,
    WorkloadCancellingError,
    WorkloadCreationError,
    WorkloadFetchingError,
)
from pasqal_cloud.job import CreateJob, Job
from pasqal_cloud.utils.constants import JobStatus  # noqa: F401
from pasqal_cloud.utils.models import (
    JobFilters,
    PaginatedResponse,
    PaginationParams,
    RebatchFilters,
)
from pasqal_cloud.workload import Workload


class SDK:
    _client: Client

    def __init__(
        self,
        username: Optional[str] = None,
        password: Optional[str] = None,
        token_provider: Optional[TokenProvider] = None,
        endpoints: Optional[Endpoints] = None,
        auth0: Optional[Auth0Conf] = None,
        webhook: Optional[str] = None,
        project_id: Optional[str] = None,
        group_id: Optional[str] = None,  # deprecated
    ):
        """
        This class provides helper methods to call the PASQAL Cloud endpoints.

        To authenticate to PASQAL Cloud, you have to provide either an
        email/password combination or a TokenProvider instance.
        You may omit the password, you will then be prompted to enter one.

        Args:
            username: Email of the user to login as.
            password: Password of the user to login as.
            token_provider: The token provider is an alternative log-in method \
              not for human users.
            webhook: Webhook where the job results are automatically sent to.
            endpoints: Endpoints targeted of the public apis.
            auth0: Auth0Config object to define the auth0 tenant to target.
            project_id: ID of the owner project of the batch.
            group_id (deprecated): Use project_id instead.
        """

        # Ticket (#622), to be removed,
        # used to avoid a breaking change during the group to project renaming
        if not project_id:
            if not group_id:
                raise TypeError(
                    "Either a 'group_id' or 'project_id' has to be given as argument"
                )
            warn(
                "The parameter 'group_id' is deprecated, from now on use"
                " 'project_id' instead",
                DeprecationWarning,
                stacklevel=2,
            )
            project_id = group_id

        self._client = Client(
            project_id=project_id,
            username=username,
            password=password,
            token_provider=token_provider,
            endpoints=endpoints,
            auth0=auth0,
        )
        self.batches: Dict[str, Batch] = {}
        self.workloads: Dict[str, Workload] = {}
        self.webhook = webhook

    def _get_batch(
        self,
        id: str,
    ) -> Dict[str, Any]:
        try:
            return self._client.get_batch(id)
        except HTTPError as e:
            raise BatchFetchingError(e) from e

    def create_batch(
        self,
        serialized_sequence: str,
        jobs: List[CreateJob],
        complete: bool = True,
        emulator: Optional[EmulatorType] = None,
        configuration: Optional[BaseConfig] = None,
        wait: bool = False,
        fetch_results: bool = False,
    ) -> Batch:
        """Create a new batch and send it to the API.

        Args:
            serialized_sequence: Serialized pulser sequence.
            jobs: List of jobs to be added to the batch at creation.
            complete: True (default), if all jobs are sent at creation.
              If set to False, jobs can be added using the `Batch.add_jobs` method.
              Once all the jobs are sent, use the `Batch.declare_complete` method.
              Otherwise, the batch will be timed out if all jobs have already
              been terminated and no new jobs are sent.
            emulator: The type of emulator to use,
              If set to None, the device_type will be set to the one
              stored in the serialized sequence
            configuration: A dictionary with extra configuration for the emulators
             that accept it.
            wait: Whether to block on this statement until all the submitted jobs are
                terminated
            fetch_results (deprecated): Whether to wait for the batch to
              be done and fetch results


        Returns:
            Batch: The new batch that has been created in the database.

        Raises:
            BatchCreationError: If batch creation failed
            BatchFetchingError: If batch fetching failed
        """
        if fetch_results:
            warn(
                "Argument `fetch_results` is deprecated and will be removed in a"
                " future version. Please use argument `wait` instead.",
                DeprecationWarning,
                stacklevel=2,
            )
            wait = wait or fetch_results

        req = {
            "sequence_builder": serialized_sequence,
            "webhook": self.webhook,
            "jobs": jobs,
            "complete": complete,
        }

        # the emulator field is only added in the case
        # an emulator job is requested otherwise it's left empty
        if emulator:
            req.update({"emulator": emulator})

        # The configuration field is only added in the case
        # it's requested
        if configuration:
            req.update({"configuration": configuration.to_dict()})  # type: ignore[dict-item]

        try:
            batch_rsp = self._client.send_batch(req)
        except HTTPError as e:
            raise BatchCreationError(e) from e

        batch = Batch(**batch_rsp, _client=self._client)

        if wait:
            while any(
                job.status in {"PENDING", "RUNNING"} for job in batch.ordered_jobs
            ):
                time.sleep(RESULT_POLLING_INTERVAL)
                batch.refresh()

        self.batches[batch.id] = batch
        return batch

    def get_batch(self, id: str, fetch_results: bool = False) -> Batch:
        """Retrieve a batch's data and all its jobs.

        Args:
            fetch_results (deprecated): Whether to wait for the batch to be
              done and fetch results
            id: ID of the batch.

        Returns:
            Batch: the batch stored in the PCS database.

        Raises:
            BatchFetchingError: in case fetching failed
        """
        if fetch_results:
            warn(
                "Argument `fetch_results` is deprecated and will be removed in a"
                " future version. Results are fetched anyway with this function.",
                DeprecationWarning,
                stacklevel=2,
            )
        batch_rsp = self._get_batch(id)
        batch = Batch(**batch_rsp, _client=self._client)
        self.batches[batch.id] = batch
        return batch

    def cancel_batch(self, id: str) -> Batch:
        """Cancel the given batch on the PCS

        Args:
            id: ID of the batch.
        """
        try:
            batch_rsp = self._client.cancel_batch(id)
        except HTTPError as e:
            raise BatchCancellingError(e) from e
        return Batch(**batch_rsp, _client=self._client)

    def rebatch(
        self,
        id: Union[UUID, str],
        filters: Optional[RebatchFilters] = None,
    ) -> Batch:
        """
        Retry a list of jobs matching filters in a new batch.

        Args:
            id: id of the batch to re-create
            filters: filters to be applied to find the jobs that will be retried

        Returns:
            An instance of a rescheduled Batch model. The fields
            can differ from the original batch as the record
            is recreated as to prevent modifying the original batch.

        Raises:
            RebatchError if rebatch call failed.
        """
        if filters is None:
            filters = RebatchFilters()
        elif not isinstance(filters, RebatchFilters):
            raise ValueError(
                f"Filters needs to be a RebatchFilters instance, not a {type(filters)}"
            )

        try:
            new_batch_data = self._client.rebatch(
                batch_id=id,
                filters=filters,
            )
        except HTTPError as e:
            raise RebatchError(e) from e
        return Batch(**new_batch_data, _client=self._client)

    def get_jobs(
        self,
        filters: Optional[JobFilters] = None,
        pagination_params: Optional[PaginationParams] = None,
    ) -> PaginatedResponse:
        """
        Retrieve a list of jobs matching filters using a pagination system.

        Jobs are sorted by their creation date in descending order.

        If no 'pagination_params' are provided, the first 100 jobs
        matching the query will be returned by default.

        Args:
            filters: Filters to be applied to the jobs.
            pagination_params: Pagination to be applied to the query.

        Examples:
        >>> get_jobs(filters=JobFilters(status=JobStatus.ERROR))
        Returns the first 100 jobs with an ERROR status.

        >>> get_jobs(filters=JobFilters(status=JobStatus.ERROR),
                     pagination_params=PaginationParams(offset=100))
        Returns jobs 101-200 with an ERROR status.

        >>> get_jobs(filters=JobFilters(status=JobStatus.ERROR),
                     pagination_params=PaginationParams(offset=200))
        Returns jobs 201-300 with an ERROR status.

        Returns:
            PaginatedResponse: A class instance with two fields:
                - total: An integer representing the total number of jobs
                         matching the filters.
                - results: A list of jobs matching the filters and pagination
                           parameters provided.

        Raises:
            JobFetchingError: If fetching jobs call failed.
            ValueError: If `filters` is not an instance of JobFilters
                    or if `pagination_params` is not an instance of PaginationParams.
        """

        if pagination_params is None:
            pagination_params = PaginationParams()
        elif not isinstance(pagination_params, PaginationParams):
            raise ValueError(
                f"Pagination parameters needs to be a PaginationParams instance, "
                f"not a {type(pagination_params)}"
            )

        if filters is None:
            filters = JobFilters()
        elif not isinstance(filters, JobFilters):
            raise ValueError(
                f"Filters needs to be a JobFilters instance, not a {type(filters)}"
            )

        try:
            response = self._client.get_jobs(
                filters=filters, pagination_params=pagination_params
            )
            jobs = response["data"]
            pagination_response = response.get("pagination")
            # It should return a pagination all the time
            total_nb_jobs = pagination_response["total"] if pagination_response else 0
        except HTTPError as e:
            raise JobFetchingError(e) from e
        return PaginatedResponse(
            total=total_nb_jobs,
            offset=pagination_params.offset,
            results=[Job(**job, _client=self._client) for job in jobs],
        )

    def _get_job(self, id: str) -> Dict[str, Any]:
        try:
            return self._client.get_job(id)
        except HTTPError as e:
            raise JobFetchingError(e) from e

    def get_job(self, id: str, wait: bool = False) -> Job:
        """Retrieve a job's data.

        Args:
            id: ID of the job.
            wait: Whether to wait for the job to be done

        Returns:
            Job: The job stored in the PCS database.

        Raises:
            JobFetchingError: If fetching failed.
        """
        job_rsp = self._get_job(id)
        if wait:
            while job_rsp["status"] in ["PENDING", "RUNNING"]:
                time.sleep(RESULT_POLLING_INTERVAL)
                job_rsp = self._get_job(id)
        return Job(**job_rsp, _client=self._client)

    def add_jobs(self, batch_id: str, jobs: List[CreateJob]) -> Batch:
        """
        Add jobs to an already existing batch.

        Args:
            batch_ID: A unique identifier for the batch data.
            Jobs: a collection of CreateJob payloads

        Returns:
            An instance of a Batch model from the PCS database

        Raises:
            JobCreationError, which spawns from a HTTPError.
            BatchAlreadyCompleteError.
        """

        try:
            resp = self._client.add_jobs(batch_id, jobs)
        except HTTPError as e:
            raise JobCreationError(e)
        return Batch(**resp, _client=self._client)

    def close_batch(self, batch_id: str) -> Batch:
        """
        Set a batch 'complete' field as True, indicating no more Jobs
        can be submitted.

        Args:
            batch_ID: A unique identifier for the batch data.

            Returns:
                An instance of a Batch model from the PCS database

        Raises:
            BatchSetCompleteError which spawns from a HTTPError
        """
        try:
            resp = self._client.complete_batch(batch_id)
        except HTTPError as e:
            raise BatchSetCompleteError(e)

        return Batch(**resp, _client=self._client)

    def cancel_job(self, id: str) -> Job:
        """Cancel the given job on the PCS

        Args:
            id: ID of the job.

        Returns:
            Job: The job stored in the PCS database.
        """
        try:
            job_rsp = self._client.cancel_job(id)
        except HTTPError as e:
            raise JobCancellingError(e) from e

        return Job(**job_rsp, _client=self._client)

    def _get_workload(self, id: str) -> Dict[str, Any]:
        try:
            return self._client.get_workload(id)
        except HTTPError as e:
            raise WorkloadFetchingError(e) from e

    def wait_for_workload(
        self, id: str, workload_rsp: Dict[str, Any]
    ) -> Dict[str, Any]:
        while workload_rsp["status"] in ["PENDING", "RUNNING"]:
            time.sleep(RESULT_POLLING_INTERVAL)
            workload_rsp = self._get_workload(id)
        return workload_rsp

    def create_workload(
        self,
        workload_type: str,
        backend: str,
        config: Dict[str, Any],
        wait: bool = False,
    ) -> Workload:
        """Create a workload to be scheduled for execution.

        Args:
            workload_type: The type of workload to create.
            backend: The backend to run the workload on.
            config: The config that defines the workload.
            wait: Whether to wait for completion to fetch results.

        Returns:
            workload: The created workload instance.

        Raises:
            WorkloadCreationError: If creation failed.

        """
        req = {
            "workload_type": workload_type,
            "backend": backend,
            "config": config,
        }
        try:
            workload_rsp = self._client.send_workload(req)
        except HTTPError as e:
            raise WorkloadCreationError(e) from e
        if wait:
            workload_rsp = self.wait_for_workload(workload_rsp["id"], workload_rsp)
        workload = Workload(**workload_rsp, _client=self._client)

        self.workloads[workload.id] = workload
        return workload

    def get_workload(self, id: str, wait: bool = False) -> Workload:
        """Retrieve a workload's data.

        Args:
            id: ID of the workload.
            wait: Whether to wait for the workload to be done.

        Returns:
            workload: The workload stored in the PCS database.

        Raises:
            WorkloadFetchingError: If fetching failed.
        """
        workload_rsp = self._get_workload(id)
        if wait:
            workload_rsp = self.wait_for_workload(id, workload_rsp)
        return Workload(**workload_rsp, _client=self._client)

    def cancel_workload(self, id: str) -> Workload:
        """Cancel the given workload on the PCS.

        Args:
            id: Workload id.

        Returns:
            workload: The canceled workload.

        Raises:
            WorkloadCancelingError: If cancelation failed.
        """
        try:
            workload_rsp = self._client.cancel_workload(id)
        except HTTPError as e:
            raise WorkloadCancellingError(e) from e
        return Workload(**workload_rsp, _client=self._client)

    def get_device_specs_dict(self) -> Dict[str, str]:
        """Retrieve the list of available device specifications.

        Returns:
            DeviceSpecs: The list of available device specifications.

        Raises:
            DeviceSpecsFetchingError: If fetching of specs failed.
        """

        try:
            return self._client.get_device_specs_dict()
        except HTTPError as e:
            raise DeviceSpecsFetchingError(e) from e
